//
// Asn Objects.
//
//
//
//
// Copyright (c) 1997-2003 Marconi Communications, Inc. as an unpublished work.
//
// All rights reserved.
//
//
//
#include <cctype>
#include <snmp/asn1/Asn1objects.h>
#include <sys/types.h>
#include <util/message_c.h>
#include <string>

#include "asn1/Asn1.h"
#include "asn1/Asn1stream.h"

using std::dec;
using std::hex;
using std::ios;
using std::ostream;
using std::streamoff;


const unsigned asnObj::local_size =  ASNOBJ_LOCAL_SIZE;
const size_t asnObj::MaxAsn1Size = 2048;
const unsigned char asnObjId::endofdata =(0x01);
const unsigned char asnObjId::endofrecord =(0x02);
const uint32_t asnObjId::read=(0);
const uint32_t asnObjId::write=(1);

//
// asn1 base object.
//
asnObj::asnObj(void) : size_(0), asn1_size_(MaxAsn1Size), type_(0) {data_.as_ptr = NULL;}

asnObj::~asnObj(void)
{
    if (size_ > local_size) {
        delete [] data_.as_ptr;
    }
}

asnObj::asnObj(const asnObj& o) : 
    size_(o.size_), 
    asn1_size_(MaxAsn1Size), 
    type_(o.type_)
{  
    // allocate the container but don't necessarily fill it in.
    data_.as_ptr = NULL;
    if (size_ > local_size) {
        data_.as_ptr = new unsigned char [size_];
	memcpy(data_.as_ptr,o.data(),size_);
    } else if (size_ > 0) {
        memcpy(data_.as_value,o.data(),size_);
    }
}

asnObj::asnObj(const unsigned char *data_in, const size_t size_in, const unsigned char type_in) :
    size_(size_in),
    asn1_size_(MaxAsn1Size),
    type_(type_in)
{
    // allocate the container but don't necessarily fill it in.
    data_.as_ptr = NULL;
    if (size_ > local_size) {
        data_.as_ptr = new unsigned char [size_];
	if (data_in) {
            memcpy(data_.as_ptr,data_in,size_);
        }
    } else if (size_ > 0) {
      if (data_in) {
          memcpy(data_.as_value,data_in,size_);
      }
    }
}

void asnObj::newValue(const unsigned char *data_in, const size_t size_in, const unsigned char type_in)
{
    // Do cleanup if new value is larger than old, or fits in as_value
    if (size_in > size_ || size_in <= local_size) {
        // delete the old value
        if (size_ > local_size) {
           delete [] data_.as_ptr;
           size_ = 0;
        }
        // allocate new array
        if (size_in > local_size) {
           data_.as_ptr = new unsigned char [size_in];
        }
    }

    // and instantiate the new value, type and size.
    size_ = size_in;
    type_ = type_in;
    asn1_size_ = MaxAsn1Size;

    if (size_ > local_size) {
       memcpy(data_.as_ptr,data_in,size_);
    } else if (size_in > 0) {
       memcpy(data_.as_value,data_in,size_);
    }
}

void asnObj::resize(const size_t newsize)
{
    if (newsize > size_ && newsize > local_size) {

        unsigned char* new_value = new unsigned char [newsize];
	memcpy(new_value, data(), size_);

	if (size_ > local_size) {
	  delete [] data_.as_ptr;
	}

	data_.as_ptr = new_value;
    }
    asn1_size_ = MaxAsn1Size;
    size_ = newsize;
}

// return the actual size in bytes of the ASN.1 BER encoding
// of this object. This function is used by get-bulk so 
// we can be smart about how many reps to resolve before
// quiting.
//
size_t asnObj::asn1EstSize(void) 
{
    if (asn1_size_ != MaxAsn1Size) {

       return asn1_size_;

    } else {

       unsigned char buffer[MaxAsn1Size];
       asn1ostream s(buffer, MaxAsn1Size);

       s << *this;

       return(asn1_size_ = s.currPos() + streamoff(1));
    }
}

//
// asn.1 object identifier constructors 
//
asnObjId::asnObjId(const unsigned char *oid, const size_t len) :
    asnObj(oid,len,ASN_OBJECT_ID),
    status(0),
    eor((uint32_t) (len/sizeof(uint32_t))),
    wrtpos((uint32_t) (len/sizeof(uint32_t))),
    rdpos(0) {}

asnObjId::asnObjId(const uint32_t *oid, const size_t len) :
    asnObj((const unsigned char*)oid,len,ASN_OBJECT_ID),
    status(0),
    eor((uint32_t) (len/sizeof(uint32_t))),
    wrtpos((uint32_t) (len/sizeof(uint32_t))),
    rdpos(0) {}

asnObjId::asnObjId(const uint32_t NrIds, const uint32_t *oid,
		   const size_t oidsize) :
    asnObj(0,NrIds*sizeof(uint32_t), ASN_OBJECT_ID),
    status(0),
    eor(NrIds),
    wrtpos(0),
    rdpos(0)
{
    uint32_t oidLen = (uint32_t) oidsize;
    
    if (NrIds) {
        if (oidLen >= NrIds) {
	    oidLen = NrIds;
	    setEndOfRecord();
	}
	if (oid && oidLen) {
	    memcpy((unsigned char*)data(), oid, oidLen * sizeof(uint32_t));
	    wrtpos = oidLen;
	}
    }
}

asnObjId::asnObjId(void) :
    asnObj(0,0,ASN_OBJECT_ID),
    status(0),
    eor(0),
    wrtpos(0),
    rdpos(0)
{ }

asnObjId::~asnObjId() { }


void asnObjId::newValue(const uint32_t *oid, const size_t size_in)
{
    asnObj::newValue((unsigned char*)oid,size_in,ASN_OBJECT_ID);
    wrtpos = eor = (uint32_t) (size_in/sizeof(uint32_t));
    rdpos = 0;
}

void asnObjId::newValue(const uint32_t NrIds,
			const uint32_t *oid, const size_t oidsize)
{
    uint32_t allocSize = (uint32_t) (NrIds*sizeof(uint32_t));
    uint32_t size_in = (uint32_t) oidsize;

    asnObj::resize(allocSize);
    if (allocSize < size_in) { size_in = allocSize; };
    eor = NrIds;
    wrtpos = size_in/sizeof(uint32_t);
    memcpy((unsigned char*)data(), oid, size_in);
}

void asnObjId::seek(const uint32_t pos, uint32_t mtype)
{
    if (mtype == read) {
        if (pos < wrtpos) {
	    rdpos = pos;
	    clrEndOfData();
	} else {
	    rdpos = wrtpos;
	    setEndOfData();
	}
    }
    if (mtype == write) {
        if (pos < eor) {
	    wrtpos = pos;
	    clrEndOfRecord();
	} else {
	    setEndOfRecord();
	    wrtpos = eor;
	}
    }
}


//
// ostream operator for printing oid objects.  
// object ids 
ostream&  operator<<(ostream& s, const asnObjId& objid)
{
    s << dec;
    for(uint32_t i=0; i<objid.ObjIdLen(); i++) {
	s << "." << (uint32_t) objid.ObjId()[i];
    }
    return s;
}

//
//  asn.1 generic asn1 print operator.
//
ostream& operator<<(ostream& s, const asnObj& o)
{
    switch(o.type()) {
       case ASN_INTEGER: {  // integer, integer32
	    if (o.size() == 1) {
		int8_t *val = (int8_t *) (o.data());
		s << (int32_t) *val;
	    } else if (o.size() == 2) {
		int16_t *val = (int16_t *) (o.data());
		s << (int32_t) *val;
	    } else {
		int32_t *val = (int32_t *) (o.data());
		s << (int32_t) *val;
	    }
	    break;
	}
        case COUNTER:  // counter, counter32
        case GAUGE:    // gauge, gauge32, unsigned32
	case TIMETICKS: {
	    if (o.size() == 8) {
		    // handle 64-bit obj
		    uint64_t *val64 = (uint64_t *) (o.data());
		    s << (uint32_t) *val64;
	    } else {
		    uint32_t *val32 = (uint32_t *) (o.data());
		    s << (uint32_t) *val32;
	    }
	    break;
	}
	case COUNTER64: {
	    uint64_t *val = (uint64_t*) (o.data());
	    s << (uint64_t) *val;
	    break;
	}
        case OPAQUE:     // opaque, bits, octetstring
        case ASN_OCTET_STR: {
	     bool inhex=false;
	     for(unsigned i=0; i<o.size(); i++) {
	       if (! isprint(o.data()[i])) {
		 inhex=true;
		 break;
	       }
	     }
	     if (inhex) {
	       // print as a list of hex characters seperated by dots.
	       s << hex;
	       for(unsigned i=0; i<o.size(); i++) {
		 s << ((uint32_t)(o.data()[i]));
		 if (i<o.size()) {
                     s << '.';
                 }
	       }
	       s << dec;
	     } else {
	       // print as a display string.
	       for(unsigned i=0; i<o.size(); i++) {
		 s << o.data()[i];
	       }
	     }
	     break;
        }
        case IPADDRESS: {  // ipaddress
	    for(unsigned i=0; i<o.size(); i++) {
	      s << (int) o.data()[i];
	    }
	    break;
	}
	case NOSUCHOBJECT: {
	    s <<  "NoSuchObject";
	    break;
	}
	case NOSUCHINSTANCE: {
	    s <<  "NoSuchInstance";
	    break;
	}
	case ENDOFMIBVIEW: {
	    s <<  "EndOfMibView";
	    break;
	}
	case ASN_NULL: {
	    s << "Null";
	    break;
	}
	case ASN_OBJECT_ID: {
	    asnObjId obj(o.data(),o.size());
	    s << (asnObjId&) obj;
	    break;
	}
	default: {
	  message_log(LOG_NOTICE,
	     "snmp: tried to write an unknown data type to standard out.");
	}
    }
    return s;
}

//
// write a asn.1 object
//
asn1ostream& operator<<(asn1ostream& s, const asnObj& o)
{
    switch(o.type()) {
        case ASN_INTEGER: {  // integer, integer32
	    s << beginPrimitive(o.type());
	    if (o.size() == 1) {
		int8_t *val = (int8_t *) (o.data());
		s << (int32_t) *val;
	    } else if (o.size() == 2) {
		int16_t *val = (int16_t *) (o.data());
		s << (int32_t) *val;
	    } else {
		int32_t *val = (int32_t *) (o.data());
		s << (int32_t) *val;
	    }
	    break;
	}
        case COUNTER:       // counter, counter32
        case GAUGE:         // gauge, gauge32
	case TIMETICKS: {
	    if (o.size() == 8) {
		    // handle 64-bit obj
		    uint64_t *val64 = (uint64_t *) (o.data());
		    s << beginPrimitive(o.type()) << (uint32_t) *val64;
	    } else {
		    uint32_t *val32 = (uint32_t *) (o.data());
		    s << beginPrimitive(o.type()) << (uint32_t) *val32;
	    }
	    break;
	}
	case COUNTER64: {
	    uint64_t *val = (uint64_t*) (o.data());
	    s << beginPrimitive(o.type())
	      << (uint64_t) *val;
	    break;
	}
        case ASN_OCTET_STR:     // octet string
        case OPAQUE:            // bits, opaque
        case IPADDRESS: {       // ipaddress
	    s << beginPrimitive(o.type())
	      << Len((int) o.size());
	    (void) s.write(o.data(),(int) o.size());
	    break;
	}
	case NOSUCHOBJECT:
	case NOSUCHINSTANCE:
	case ENDOFMIBVIEW:
	case ASN_NULL: {
	    s << asnNull(o.type());
	    break;
	}
	case ASN_OBJECT_ID: {
	    asnObjId obj(o.data(),o.size());
	    s << obj;
	    break;
	}
	default: {
	  message_log(LOG_ERR,
	     "snmp: tried to write an unknown data type to ASN.1 stream.");
	}
    }
    return s;
}

//
// For Encoding Oids
//
const int32_t maxOidSiz(128);

static void parse_asn1_objidentifier(asn1istream& s,uint32_t objId[],uint32_t* len)
{
    uint32_t maxOidSize = *len;
    uint32_t subidentifier = 0;
    ssize_t numbytes=0;
    unsigned char c=0;

    s >> c;
    if ((unsigned char)c != ASN_OBJECT_ID) {
	s.setstate(ios::badbit|ios::failbit);
    } else {
	if ((numbytes = getLength(s)) > 0) {
	    /* account for expansion of first byte */
	    uint32_t i=1;
	    for(; numbytes && i<maxOidSize; i++) {
		subidentifier = 0;
		do {
		    s>>c;
		    /* shift and add in low order 7 bits */
		    subidentifier= (subidentifier<<7)+((unsigned char)c&~ASN_BIT8);
		    numbytes--;
		    /* last byte has high bit clear */
		} while (((unsigned char)c & ASN_BIT8)&&numbytes);

		objId[i]=subidentifier;
	    }
	    if (!s.eof()) {
		// Finish off oid.
		/*
		 * The first two subidentifiers are encoded into the first
		 * component with the value (X * 40) + Y, where:
		 * X is the value of the first subidentifier.
		 * Y is the value of the second subidentifier.
	         */
		 if (objId[1] == 0x2B){
		     objId[0] = 1;
		     objId[1] = 3;
		 } else if (objId[1] < 40) {
                     objId[0] = 0;
                 } else if (objId[1] < 80) {
                     objId[0] = 1;
                     objId[1] -= 40;
                 } else {
                     objId[0] = 2;
                     objId[1] -= 80;
                 }

		 *len = i;
	    }
	} else {
	    // Handle invalid object identifier encodings
	    // of the form 06 00 robustly.
	    objId[0] = objId[1] = 0;
	    *len = 2;
	}
    }
}



//
// read an asn object from the asn1 data stream
//
asn1istream& operator>>(asn1istream& s, asnObj& o)
{
    unsigned char buf[2048];
    unsigned char asn1Type=0;
    size_t numbytes;
    int32_t bytesRead;

    if (s.good()) {
	s >> asn1Type;
	if (s.good()) {
          switch(asn1Type) {
	    case ASN_INTEGER: {  // integer, integer32
		int32_t i;
		s >> i;
		if (s.good()) {
		  o.newValue((unsigned char*)&i,sizeof(int32_t),asn1Type);
		}
		break;
	    }
     	    case COUNTER:        // counter, counter32
  	    case GAUGE:          // gauge, gauge32, unsigned
	    case TIMETICKS: {
		uint32_t i;
		s >> i;
		if (s.good()) {
		  o.newValue((unsigned char*)&i,sizeof(uint32_t),asn1Type);
		}
		break;
	    }
	    case COUNTER64: {
		uint64_t i;
		s >> i;
		if (s.good()) {
		  o.newValue((unsigned char*)&i,sizeof(uint64_t),asn1Type);
		}
		break;
	    }
 	    case ASN_OCTET_STR:  // octet string, bits, opaque
	    case IPADDRESS: {    // ipaddress
		if ((numbytes = getLength(s)) > 0) {
		    bytesRead = s.read(buf,2048,(unsigned int) numbytes);
		    if (s.good()) {
			o.newValue(buf,bytesRead,asn1Type);
		    }
		} else {
		    o.newValue(0,0,asn1Type);
		}
		break;
	    }
	    case ASN_OBJECT_ID: {
		uint32_t len = maxOidSiz;
		uint32_t objId[maxOidSiz];
    
		s.unget(asn1Type);
		parse_asn1_objidentifier(s, objId, &len);
		if (s.good()) {
		  o.newValue((unsigned char*)objId,len*sizeof(uint32_t),asn1Type);
		}
                else {
	          s.setstate(ios::badbit|ios::failbit);
                }
		break;
	    }
	    case NOSUCHOBJECT: 
  	    case NOSUCHINSTANCE:
            case ENDOFMIBVIEW: 
	    case ASN_NULL: {
		unsigned char c;
		s >> c;
		if (s.good()) {
		  o.newValue(0,0,asn1Type);
		}
		break;
	    }
	    default: {
	      s.setstate(ios::badbit|ios::failbit);
	    }
          }
	} else {
	    s.setstate(ios::badbit|ios::failbit);
        }

    }
    return s;
}


//
//  encode an object identifier in asn.1 and write it
//  to the output stream.
//
asn1ostream& operator<<(asn1ostream& s, const asnObjId& objid)
{
    unsigned char    buf[maxOidSiz * (sizeof(uint32_t) + 1)];
    unsigned char    *bp = buf;
    uint32_t *op = objid.ObjId();
    int32_t   objLength = objid.ObjIdLen();
    uint32_t subid, mask, testmask;
    int32_t   bits, testbits, numbytes;
    
	
    if (objLength < 2) {
	*bp++ = (unsigned char) 0;
	objLength = 0;
    } else {
        op[1] += (op[0] * 40);
        op++;
        objLength--;  
    }

    while(objLength-- > 0){
	subid = *op++;
	if (subid < 0x80){ 
	    *bp++ = (unsigned char) subid;
	} else {
	    mask = ((uint32_t) 0x7F); /* handle subid == 0 case */
	    bits = 0;
	    /* testmask *MUST* !!!! be of an unsigned type */
	    for(testmask = 0x7F, testbits = 0; testmask != 0;
		testmask <<= 7, testbits += 7){
		if (subid & testmask){	/* if any bits set */
		    mask = testmask;
		    bits = testbits;
		}
	    }
	    /* mask can't be zero here */
	    for(;mask != 0x7F; mask >>= 7, bits -= 7){
		/* fix a mask that got truncated above */
                if (mask == 0x1E00000) {
		    mask = 0xFE00000;
                }
		*bp++ = (unsigned char)(((subid & mask) >> bits) | ASN_BIT8);
	    }
	    *bp++ = (unsigned char)(subid & mask);
	}
    }
    numbytes = (int32_t) (bp-buf);
    
    s << beginPrimitive(ASN_OBJECT_ID) << Len(numbytes);
    (void) s.write(buf, numbytes);
    return s;
}

//
//  read an object identifier from the asn.1 stream.
//
asn1istream& operator>>(asn1istream& s, asnObjId& objid)
{
    uint32_t len = maxOidSiz;
    uint32_t objId[maxOidSiz];
    
    parse_asn1_objidentifier(s, objId, &len);

    if (s.good()) {
      objid.newValue(objId,len*sizeof(uint32_t));
    }
    else {
       s.setstate(ios::badbit|ios::failbit);
    }
    return s;
}

