// Emacs editing mode: -*- C++ -*- 
//
// asn1 container objects 
//
//
//
// Copyright (c) 1997-2003 Marconi Communications, Inc. as an unpublished work.
//
// All rights reserved.
//
//
//
//
#ifndef _asn1objects_h_
#define _asn1objects_h_
#include <assert.h>
#include <snmp/asn1/Asn1.h>
#include <snmp/asn1/Asn1stream.h>
#include <stddef.h>
#include <stdint.h>
#include <string.h>
#include <ostream>

#define ASNOBJ_LOCAL_SIZE 8

//
//  asn.1 container class.
//
//  Note: 60-80% of all values (excluding obj-ids) carried in
//  this container are 8-bytes or less so as an optimization values
//  less than 8-bytes will be stored in this container
//  instead of allocating a separate object on the heap.
//
class asnObj {
    static const unsigned local_size;
    static const size_t MaxAsn1Size;    
    union {
        unsigned char *as_ptr;
        unsigned char as_value[ASNOBJ_LOCAL_SIZE];
    } data_;

    size_t size_;
    size_t asn1_size_;
    unsigned char type_;
public:
    asnObj(const unsigned char *value, const size_t size, const unsigned char type);
    asnObj(const asnObj&);
    asnObj(void);
    virtual ~asnObj(void);

    // assign a new-value to this object.
    void newValue(const unsigned char *value, const size_t length, const unsigned char type);

protected:
    // resize - increase the size of this object.
    void resize(const size_t newsize);

public:
    // assign a error value to this object.
    void noSuchObject(void) {
        newValue(0,0,NOSUCHOBJECT);
    };

    void noSuchInstance(void) {
        newValue(0,0,NOSUCHINSTANCE);
    };

    void endOfMibView(void) {
        newValue(0,0,ENDOFMIBVIEW);
    };

    bool isEndOfMibView(void) {
        return (type_ == ENDOFMIBVIEW);
    };

    void null(void) {
        newValue(0,0,ASN_NULL);
    };

    bool isUndefined(void) {
        return (type_ == 0);
    };
    
    bool isNoSuchInstance() {
        return (type_ == NOSUCHINSTANCE);
    }

    size_t asn1EstSize(void);

    size_t size(void) const { return size_; };

    unsigned char type(void) const { return type_; };

    unsigned char* data(void) const {
        return (size_ <= local_size) ? (unsigned char*) data_.as_value : data_.as_ptr;
    }

    // i/o operators for reading and writing asn.1 
    friend asn1ostream& operator<<(asn1ostream& s, const asnObj& o);
    friend asn1istream& operator>>(asn1istream& s, asnObj& o);

    // i/o for writing to a ostream
    friend std::ostream& operator<<(std::ostream& s, const asnObj& o);
};    

//
//  ASN.1 integer base class
//
template<class T, unsigned char TYPE> 
class asnIntObj : public asnObj {
public:
    asnIntObj(T v) : asnObj((unsigned char*)&v, sizeof(T), TYPE) {};
    asnIntObj(void) : asnObj(0, 0, TYPE) {};
    T data(void) const { return (T) (* ((T*) asnObj::data())); };
    
    friend asn1ostream& operator<<(asn1ostream& s, const asnIntObj& o) {
        extern asn1ostream& _beginPrimitive(asn1ostream& s, unsigned char t);
	return s << beginPrimitive(TYPE) << (T) o.data();
    };
    
    friend asn1istream& operator>>(asn1istream& s, asnIntObj& o) {
	s >> (asnObj&) o;
	return s;
    };

    friend std::ostream& operator<<(std::ostream& s, const asnIntObj& o) {
      return s << " Int(" << std::dec << (T) o.data() << " ) ";
    };
};


//
//  ASN.1 SMIv2 numerical types.
//
typedef asnIntObj<int32_t,ASN_INTEGER>   asnInteger32;
typedef asnIntObj<uint32_t,UNSIGNED32>  asnUnsigned32;
typedef asnIntObj<uint32_t,COUNTER>     asnCounter32;
typedef asnIntObj<uint32_t,GAUGE>       asnGauge32;
typedef asnIntObj<uint32_t,TIMETICKS>   asnTimeTicks;
typedef asnIntObj<uint64_t,COUNTER64>   asnCounter64;


//
//  ASN SMIv2 string base classes
//
template<class T, unsigned char TYPE> 
class asnStrObj : public asnObj {
public:
    asnStrObj(void) : asnObj(0,0,TYPE) {};
    asnStrObj(const char *o) :
      asnObj((const T*)o, strlen(o), TYPE) {};
    asnStrObj(const T* o, const size_t l) : asnObj(o,l,TYPE) {};
    asnStrObj(const char* o, const size_t l) :
      asnObj((const T*)o,l,TYPE) {};
    ~asnStrObj() { };

    void newValue(T* o, size_t l) { asnObj::newValue((unsigned char*)o, l, type()); };

    friend asn1ostream& operator<<(asn1ostream& s, const asnStrObj& o) {
	assert(o.type() == TYPE);
	if (o.type() == TYPE) {
	  s << (const asnObj&) o;
	} else {
	  s.setstate(std::ios::failbit|std::ios::badbit);
	}
	return s;
    };
    
    friend asn1istream& operator>>(asn1istream& s, asnStrObj& o) {
	s >> (asnObj&) o;
	if (o.type() != TYPE) {
            s.setstate(std::ios::failbit|std::ios::badbit);
        }
	return s;
    };

    friend std::ostream& operator<<(std::ostream& s, const asnStrObj& o) {
      s << "str(" << std::dec << o.size() << ") " << std::hex ;
       for(uint32_t i=0; i<o.size(); i++) {
	   s << "." << (uint32_t) o.data()[i];
       }
       return s;
    };
    T *data(void) const { return ((T*) asnObj::data()); };    
};


//
//  ASN.1 SMIv2 string types.
//  (note NSAP and Bit String have been removed from the
//   SMIv2 and they are now historic).
//
typedef asnStrObj<unsigned char, ASN_OCTET_STR> asnOctetStr;
typedef asnStrObj<unsigned char, IPADDRESS>     asnIpAddress;

//
//  ASN.1 SMIv2 object identifier type.
//
class asnObjId : public asnObj {
    static const unsigned char endofdata;
    static const unsigned char endofrecord;
    unsigned char status;                   // status of the last read/write operation.
    uint32_t eor;                   // end of the allocated space.
    uint32_t wrtpos;
    uint32_t rdpos;
  
    void setEndOfData(void) { status |= endofdata; };
    void clrEndOfData(void) { status &= ~endofdata; };
  
    void setEndOfRecord(void) { status |= endofrecord; };
    void clrEndOfRecord(void) { status &= ~endofrecord; };
public:

    // the arguments to this constructor are:
    // oid - pointer to an array on longs
    // len - the length of that array in size_t.
    // ex ... asnObjId(foo, sizeof(foo))
    //
    asnObjId(const uint32_t *oid, const size_t len);
    asnObjId(const unsigned char *oid, const size_t len);
    asnObjId(const uint32_t NrIds, const uint32_t *oid=0,
	     const size_t oidsize=0);
    asnObjId(void);
    ~asnObjId(void);

    void newValue(const uint32_t *oid, const size_t oidsize);
    void newValue(const uint32_t NrIds, const uint32_t *oid,
		  const size_t size);
    size_t size(void) const { return wrtpos * sizeof(uint32_t); };

    // asn i/o operators
    friend asn1ostream& operator<<(asn1ostream&, const asnObjId& );
    friend asn1istream& operator>>(asn1istream&, asnObjId& );
    
    friend std::ostream&  operator<<(std::ostream& s, const asnObjId& objid);
    friend std::istream&  operator>>(std::istream& s, asnObjId& objid);

    // copy this Oid object and return a pointer to the copy.
    asnObjId* copy() {
      asnObjId* o = new asnObjId();
      if (o) {
          o->newValue(ObjId(), ObjIdLen());
      }
      return o;
    };

    // get object identifier.
    uint32_t* ObjId(void) const { return (uint32_t*) asnObj::data(); };
    uint32_t ObjIdLen(void) const { return wrtpos; };

    // get a sub-identifier
    uint32_t readSubId(void) {
      if (rdpos < wrtpos) {
	return ObjId()[rdpos++];
      } else {
	setEndOfData();
	return 0;
      }
    };

    // move the read/write marker
    // to a position within the object-identifier.
    // args:
    //       position: that the marker will be placed.
    //       optype: read(0), write(1)
    static const uint32_t read, write;

    uint32_t curReadPos(void) { return rdpos; };

    void seek(const uint32_t pos, uint32_t mtype);

    // failure tests
    bool EndOfData() { return (status & endofdata); };
    bool EndOfRecord() { return (status & endofrecord); };

    // set a sub-identifier
    // (the eod marker is moved to the postion after id.
    bool wrtSubId(uint32_t val) {
        if (wrtpos < eor) {
	   ObjId()[wrtpos++] = val;
	   return true;
	} else {
	   setEndOfRecord();
	   return false;
	}
    };
};
#endif /* _asn1objects_h_ */
