//
//
//
// Copyright (c) 1997-2003 Marconi Communications, Inc. as an unpublished work.
//
// All rights reserved.
//
//
//
// Copyright (C) 1997 Scalable Networks, Inc.  All rights reserved.
//
//  This file implements ASN.1 C++ stream based encoding
//  and decoding. Much of the asn encoding and decoding 
//  code is from CMU SNMP snmplib/asn1.c. 
//
/***********************************************************
	Copyright 1988, 1989, 1990 by Carnegie Mellon University
	Copyright 1989	TGV, Incorporated

		      All Rights Reserved

Permission to use, copy, modify, and distribute this software and its
documentation for any purpose and without fee is hereby granted,
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in
supporting documentation, and that the name of CMU and TGV not be used
in advertising or publicity pertaining to distribution of the software
without specific, written prior permission.

CMU AND TGV DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
EVENT SHALL CMU OR TGV BE LIABLE FOR ANY SPECIAL, INDIRECT OR
CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF
USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
PERFORMANCE OF THIS SOFTWARE.
******************************************************************/
//
#include <cassert>
#include <snmp/asn1/Asn1stream.h>


#include "asn1/Asn1.h"

using std::ios;
using std::streambuf;
using std::streampos;

//
//  Constructors
//

asn1ostream::asn1ostream(int len) : streambuf(), asn1ObjContext(), contextIndex(0), encrypt_(false)
{
  ios::init(this);
  m_stringBuff  = new char [len];
  if (m_stringBuff)
    {
      setp((char *)m_stringBuff, (char *)(m_stringBuff + len));
    }
}

asn1ostream::asn1ostream(unsigned char *b, size_t len) : streambuf(),
         asn1ObjContext(), contextIndex(0), encrypt_(false)
{
  ios::init(this);
  setp((char*)b, (char*)(b+len));
  m_stringBuff = NULL;
}

asn1ostream::~asn1ostream(void)
{
	if (m_stringBuff != NULL ) {
		delete [] m_stringBuff; m_stringBuff = NULL;
	}
}

//
// asn1istream constructors
//

// passed buffer
asn1istream::asn1istream(unsigned char *b, int32_t len) :streambuf(),
                              asn1ObjContext(), contextIndex(0), currRecLength(), recIndex(0)
{
  ios::init(this);
  setg((char *)b, (char *) b, (char *)(b + len));
}

//
// asn1istream (input) operators
//
//
asn1ostream& _endObj(asn1ostream& s);
size_t getLength(asn1istream& s);

//
//  begin an asn1 object of type 'X'
//  Don't save space for the length bytes. Each primitive
//  will take care of it on it's own.
// 
asn1ostream& _beginPrimitive(asn1ostream& s, unsigned char type)
{
    return s << (unsigned char)(ASN_PRIMITIVE|type);
}

//  Fill the length field with a bogus length as 
//  a place holder.
asn1ostream& _beginContext(asn1ostream& s, unsigned char type)
{
    const unsigned char c(0);
    s.pushASNContext(s.currPos());
    s << (unsigned char)(ASN_CONTEXT | ASN_CONSTRUCTOR | type);
    return s << c << c << c;
}

//
//  endContext() - finishes the asn.1 encoding of the pdu
//  implied sequence by determining the length of the
//  newly encoded object and writing that length into the
//  ans.1 length field. For implied sequences allow the use 
//  the short form length encoding.
//
asn1ostream& endContext(asn1ostream& s)
{
    streampos oldpos = s.length();
    
    // go the the start of this sequence plus
    // one byte to skip over the tag.
    streampos pos = s.popASNContext() + streampos(1);
    (void) s.seek(pos);

    // calculate the object length by subtracting
    // the end of the object from the beginning less
    // the length bytes.
    int32_t objLength = (int32_t) (oldpos - pos - streampos(ASN_LONGLENGTH_SIZE));
    assert(objLength >= 0);
    
    if (objLength < 127) {
        // encode the short form length to make the
        // ancient version of the powerhub ILMI happy.
        s << (unsigned char)(objLength & 0x7F);
        streampos cpos = s.currPos();
        s.move(cpos,cpos+streampos(2),oldpos);
        (void) s.seek(oldpos-streampos(2));
    } else {
        s << (unsigned char)(0x02 | ASN_LONG_LEN);
	s << (unsigned char)(((uint32_t) objLength >> 8) & 0xFF);
	s << (unsigned char)(objLength & 0xFF);
	(void) s.seek(oldpos);
    }
    return s;
}

//
// encode a null
//
// (not necessarily an ASN_NULL.
// SNMP_NOSUCHOBJECT, SNMP_NOSUCHINSTANCE, SNMP_ENDOFMIBVIEW
// shipped this way too).
// 
//
asn1ostream& _null(asn1ostream& s, unsigned char type)
{
    return s << type << Len(0);
}

asn1ostream& beginSequence(asn1ostream& s)
{
    const unsigned char c(0);
    s.pushASNContext(s.currPos());
    s << (unsigned char) (ASN_SEQUENCE | ASN_CONSTRUCTOR);
    return s << c << c << c;
}

asn1ostream& endSequence(asn1ostream& s)
{
    return s << _endObj;
}

asn1ostream& beginSecurityParams(asn1ostream& s)
{
    const unsigned char c(0);
    s.pushASNContext(s.currPos());
    s << (unsigned char) (ASN_OCTET_STR);
    return s << c;
}

asn1ostream& endSecurityParams(asn1ostream& s)
{
    int32_t cpos = s.length();
    int32_t pos = (int32_t) (s.popASNContext() + streampos(1));
    (void) s.seek(pos);

    // calculate the object length by subtracting
    // the end of the object from the beginning less
    // the length bytes.
    //
    int32_t objLength = cpos - pos -1;

    assert(objLength >= 0);
    s << (unsigned char)objLength;
    (void) s.seek(cpos);
    return s;
}

//
//  _cancelObj -  cancel object causes the previously
//  written sequence or context to be canceled.
//  
asn1ostream& _cancelObj(asn1ostream& s)
{
    int32_t pos = (int32_t) s.popASNContext();
    (void) s.seek((pos == EOF) ? s.length() : pos);
    s.ios::setstate(ios::badbit|ios::failbit);
    return s;
}

asn1ostream& cancelSequence(asn1ostream& s)
{
    return _cancelObj(s);
}

//
// readSequence - read an asn.1 sequence header. Save
//                the current context and object length.
//
asn1istream& readSequence(asn1istream& s)
{
    unsigned char c=0;
    s >> c;
    // parse seq type and the length of the sequence.
    if (!s.eof() && ((unsigned char)(ASN_SEQUENCE | ASN_CONSTRUCTOR) == c)) {
	s.pushASNSequenceLen(getLength(s));
	s.pushASNContext(s.currPos());
    } else {
	s.ios::setstate(ios::badbit);
    }
    return s;
}

asn1istream& readSecurityParams(asn1istream& s)
{
    unsigned char c=0;
    s >> c;
    // parse seq type and the length of the sequence.
    if (!s.eof() && ((unsigned char)(ASN_OCTET_STR) == c)) {
        // s.pushASNSequenceLen(getLength(s));
	// consume the next length character
	// we are avoiding the check for the right security blob length
	s >> c;
    } else {
        s.ios::setstate(ios::badbit);
    }
    return s;
}

//
//  readContext - read an asn.1 implied sequence header.
//                (Only PDU types are implied.) Save the current
//                context and object length.
//
int32_t readContext(asn1istream& s)
{
    unsigned char c=0;
    s >> c;
    if (!s.eof() && (((unsigned char)(ASN_CONTEXT|ASN_CONSTRUCTOR)&c) ==
	(unsigned char)(ASN_CONTEXT|ASN_CONSTRUCTOR))) {
	// consume object length bytes
        uint32_t len = (uint32_t) getLength(s);
	if (len == 0) {
	    s.ios::setstate(ios::badbit);
	    return (EOF);
	} else {
	    s.pushASNSequenceLen(len);
	    s.pushASNContext(s.currPos());
	    return (0x01f & c);
	}
    } else {
	s.ios::setstate(ios::badbit);
	return (EOF);
    }
}

// _endObj() -
//
// At the end of an object pop the current context off
// of the stack and calculate the bytes consumed.
// If the bytes consumed don't match the number
// of bytes read in the sequence lenght field
// the parse is bad.
//
asn1istream& _endObj(asn1istream& s)
{
    size_t length = s.currPos() - s.popASNContext();
    size_t recordLength = s.popASNSequenceLen();
    if (length != recordLength) {
	s.ios::setstate(ios::badbit);
    }
    return s;
}

//  isSequenceEnd() - determine whether or
//  not a sequence has been consumed.
//
bool isSequenceEnd(asn1istream& s)
{
    size_t length = s.currPos() - s.peekASNContext();
    size_t recordLength = s.peekASNSequenceLen();
    return (length == recordLength);
}

asn1istream& endSequence(asn1istream& s)
{
    return s >> _endObj;
}

asn1istream& endContext(asn1istream& s)
{
  return s >> _endObj;
}

//
// _endObj() -
//
// At the end of the object pop the context off
// of the stack and seek to that position in the buffer.
// Calculate the length and update the length of the 
// object and asn encode it after the sequence tag
// and before the object value.
//
asn1ostream& _endObj(asn1ostream& s)
{
    int32_t cpos = s.length();
    int32_t pos = (int32_t) (s.popASNContext() + streampos(1));
    (void) s.seek(pos); 
  
    // calculate the object length by subtracting
    // the end of the object from the beginning less
    // the length bytes.
    //
    int32_t objLength = cpos - pos - ASN_LONGLENGTH_SIZE;
    
    assert(objLength >= 0);
    s << (unsigned char)(0x02 | ASN_LONG_LEN);
    s << (unsigned char)(((uint32_t) objLength >> 8) & 0xFF);
    s << (unsigned char)(objLength & 0xFF);
    (void) s.seek(cpos);
    return s;
}

//
//  _Len - asn encode and write the length parameter.
//
//         Assume that we will never encode an object
//         bigger then 64kb.
//
asn1ostream& _Len(asn1ostream& s, int32_t length) 
{
    if (length < 0x80) {
	s << (unsigned char) length;
    } else if (length < 0xFF) {
	s << (unsigned char)(0x01|ASN_LONG_LEN);
	s << (unsigned char) length;
    } else {
	s << (unsigned char)(0x02 | ASN_LONG_LEN);
	s << (unsigned char)(((uint32_t) length >> 8) & 0xFF);
	s << (unsigned char)(length & 0xFF);
    }
    return s;
}

//
//  getLength - parse and return the length of an ASN1 object.
//
size_t getLength(asn1istream& s) 
{
    int32_t lengthbyte;
    int32_t value=0;
    unsigned char c=0;

    s >> c;
    if (!s.eof() && (c & ASN_LONG_LEN)) {
	c &= ~ASN_LONG_LEN;	/* turn MSb off */
	if (c == (unsigned char)0){
	    s.ios::setstate(ios::badbit);
	}
	if (c > sizeof(uint32_t)){
	    s.ios::setstate(ios::badbit);
	}
	lengthbyte=(int)c;
	while(lengthbyte--) {
	  s >> c;
	  value = (value << 8) | c;
	}
    } else { /* short asnlength */
	value =  c; 
    }
    return (s.good()) ? value : 0;
}

//
//  reset - reset the state of the stream buffer.
//
//
void asn1istream::reset(void) 
{
  ios::clear();
  streambuf::setg(streambuf::eback(), streambuf::eback(), streambuf::egptr());
}

void asn1ostream::reset(void) 
{
  ios::clear();
  streambuf::setp(streambuf::pbase(), streambuf::epptr());
}


//
//
// I/O functions for writing asn1 to a buffer.
//
//  Integer
//
asn1ostream& asn1ostream::operator<<(int32_t integer)
{
    unsigned char buf[sizeof(int)];
    unsigned char *bp = buf;
    uint32_t mask, intsize = sizeof(int);

    mask = ((uint32_t) 0x1FF) << ((8 * (sizeof(int) - 1)) - 1);
    /* mask is 0xFF800000 on a big-endian machine */

    /*
     * Strip "unnecessary" bytes off of the most significant end of this
     * 2's complement integer.  There should be no sequence of 9
     * consecutive 1's or 0's at the most significant end of the
     * integer.
     */
    while((((integer & mask) == 0) || ((integer & mask) == mask))
	  && intsize > 1){
	intsize--;
	integer <<= 8;
    }

    /* mask is 0xFF000000 on a big-endian machine */
    mask = ((uint32_t) 0xFF) << (8 * (sizeof(uint32_t) - 1));

    // Now copy the significant bytes to the asn1 stream.
    while(intsize--){
	*bp++ = (unsigned char)((integer & mask) >> (8 * (sizeof(uint32_t) - 1)));
	integer <<= 8;
    }
    intsize = (uint32_t) (bp - buf);
    (*this) << Len(intsize);
    (void) this->write(buf, intsize);
    
    return (*this);
}

//
// called to read the integer. Type should already
// have been read. Just need to parse the length
// and read the actual number.
//
asn1istream& asn1istream::operator>>(int32_t& integer)
{
    uint32_t numbytes=0;
    int32_t value=0;
    unsigned char c=0;


    if ((numbytes = (uint32_t) getLength((*this))) > 0) {
	if (numbytes <= sizeof(int)) {
	    (*this) >> c;
	    if (c & 0x80) { // integer is negative 
		value = -1;
	    }
	    value = (value << 8) | c;
	    for(numbytes--; numbytes && !eof(); numbytes--) {
		(*this)>>c;
		value = (value << 8) | c;
	    }
	    if (! numbytes) {
	      integer = value;
	    } else {
	      this->ios::setstate(ios::badbit|ios::failbit);
	    }
	} else {
	  this->ios::setstate(ios::badbit);
	}
    }
    return (*this);
}

//
//  Unsigned Integer
//  
asn1ostream& asn1ostream::operator<<(uint32_t integer)
{
    return (*this) << (uint64_t) integer;
}

asn1istream& asn1istream::operator>>(uint32_t& integer)
{
    uint64_t value;
    (*this) >> value;
    integer = (uint32_t) value;
    return (*this);
}

//
// and now uint64_t 
//
asn1ostream& asn1ostream::operator<<(uint64_t integer)
{
    int32_t add_null_byte = 0;
    unsigned char buf[sizeof(uint64_t)]; // max number of
    unsigned char *bp = buf;
    
    uint64_t mask;
    int32_t intsize = sizeof(uint64_t);

    mask = ((uint64_t) 0xFF) << (8 * (sizeof(uint64_t) - 1));
    /* mask is 0xFF000000 on a big-endian machine */

    // if the "most significant bit" is set then...
    if ((unsigned char)((integer & mask) >> (8 * (sizeof(uint64_t) - 1))) & 0x80){
	add_null_byte = 1;
	intsize--;
    }

    /* mask is 0xFF800000 on a big-endian machine */
    mask = ((uint64_t) 0x1FF) << ((8 * (sizeof(uint64_t) - 1)) - 1);
    
    
    /*
     * Strip "unnecessary" bytes off of the most significant end of this
     * 2's complement integer. There should be no sequence of 9
     * consecutive 1's or 0's at the most significant end of the
     * integer.
     */
    while((((integer & mask) == 0) || ((integer & mask) == mask))
	  && intsize > 1){
	intsize--;
	integer <<= 8;
    }

    if (add_null_byte == 1){
	*bp++ = '\0';
	intsize--;
    }

    // now copy the rest of the number
    mask = ((uint64_t) 0xFF) << (8 * (sizeof(uint64_t) - 1));
    /* mask is 0xFF000000 on a big-endian machine */
    while(intsize--){
	*bp++ = (unsigned char)((integer & mask) >> (8 * (sizeof(uint64_t) - 1)));
	integer <<= 8;
    }
    intsize = (int32_t) (bp - buf);
    (*this) << Len(intsize);
    (void) this->write(buf,intsize);
    return (*this);
}

asn1istream& asn1istream::operator>>(uint64_t& integer)
{
    uint64_t value=0;
    uint32_t numbytes=0;

    if ((numbytes = (uint32_t) getLength((*this))) != 0) {
	// account for extra null byte when high order bit is set
	if (numbytes <= sizeof(uint64_t)+1) {
	  for(; numbytes && !eof(); numbytes--) {
	    unsigned char c=0;
	    (*this) >> c;
	    value = (value<<8) | c;
	  }
	  if (! numbytes) {
	    integer = value;
	  } else {
	    this->ios::setstate(ios::badbit|ios::failbit);
	  }
	} else {
	    this->ios::setstate(ios::badbit);
	}
    }
    return (*this);
}

//  
// output a null terminated display string.
//
asn1ostream& asn1ostream::operator<<(char* s)
{
    int32_t length_in = (int32_t) (strlen(s)+1);
    (*this) << beginPrimitive(ASN_OCTET_STR) << Len(length_in);
    while(*s) {
        this->put(*s++);
    }
    this->put(*s); // don't forget the null char.
    return (*this);
}

//
// read a string from the stream
// (numbytes is read from the asn1 input stream)
//
int32_t asn1istream::read(unsigned char *cp, size_t len, uint32_t NrBytesWanted)
{
  uint32_t NrRead=0;

  if (good() && NrBytesWanted > 0) {
      // copy bytes
      while(NrRead<NrBytesWanted && NrRead < len) {
	  (*this) >> cp[NrRead];
	  if (eof()) {
	      break;
	  }
	  NrRead++;
      }
  }
  return NrRead;
}

//
// write a buffer back into the asn1istream
//
asn1istream& asn1istream::rewrite(const char *data_in, int32_t len)
{
    // The output pointers must be set appropriately for sputc
    if (streambuf::epptr() <= streambuf::pptr()) {
	streambuf::setp(streambuf::eback(), streambuf::egptr());
    }
    while(len-- > 0 && this->good()) {
	if (sputc((int)*data_in++) == EOF) {
	    ios::setstate(ios::badbit|ios::failbit);
	}
    }
    return (*this);
}
