//
//  asn1stream module for implementing a serialization buffer
//  for encoding and decoding asn1 streams.
//
//
//
// Copyright (c) 1997-2003 Marconi Communications, Inc. as an unpublished work.
//
// All rights reserved.
//
//
//
//
#ifndef _asn1stream_h_
#define _asn1stream_h_

#include <libntaperror/ref.h>
#include <stdint.h>
#include <stdio.h>
#include <string.h>
#include <iostream>
#include <string>


//
//  asn1 (output) stream
//
class asn1ostream : public std::streambuf,
		    public virtual std::ios, public refc<asn1ostream>
{
    // a position pointer remembers a point in
    // the stream buffer. It is used in the asn1ObjContext
    // stack to record the starting points of ASN1
    // context and sequences.
    std::streampos asn1ObjContext[10];
    int32_t contextIndex;

    void pushASNContext(std::streampos p) {
        if (contextIndex < 10) {
	    asn1ObjContext[contextIndex] = p;
	    contextIndex++;
	} else {
	    std::ios::setstate(std::ios::badbit|std::ios::failbit);
	}
    }

    std::streampos popASNContext(void) {
      if (contextIndex > 0) {
	  contextIndex--;
	  return asn1ObjContext[contextIndex];
      } else {
	  std::ios::setstate(std::ios::badbit|std::ios::failbit);
      }
      return 0;
    }

    void put(unsigned char c) {
        if (sputc((int) c) == EOF) {
	    std::ios::setstate(std::ios::badbit|std::ios::failbit);
	}
    }

    virtual int	overflow(int c=EOF) { return EOF; }

    int blen() const { return (int) (epptr() - pbase()); }

public:

    // get current position of the write pointer
    std::streampos currPos(void) {
        std::streampos offset = pptr() - pbase();
        return (offset >= blen()) ? std::streampos(EOF) : offset;
    };

    friend asn1ostream& _endObj(asn1ostream&);
    friend asn1ostream& endContext(asn1ostream&);
    friend asn1ostream& _cancelObj(asn1ostream&);
    friend asn1ostream& beginSequence(asn1ostream&);
    friend asn1ostream& beginSecurityParams(asn1ostream&);
    friend asn1ostream& endSecurityParams(asn1ostream&);
    friend asn1ostream& _beginPrimitive(asn1ostream& s, unsigned char type);
    friend asn1ostream& _beginContext(asn1ostream& s, unsigned char type);

    // Constructors
    // default buffer size
    // write sequence header next?

    // allocate a "static" buffer of a given size "len".
    asn1ostream(int len);

    // use previously allocated buffer "b" of size "len".
    asn1ostream(unsigned char *b, size_t len);

    // Destructor
    ~asn1ostream(void);

    int32_t seek(std::streampos newpos) {
          char *my_pbase, *my_epptr;
	  my_pbase = pbase();
	  my_epptr = epptr();
	  if (newpos > blen()) {
	     std::ios::setstate(std::ios::badbit|std::ios::failbit);
	     return EOF;
	  } else {
	     setp(my_pbase, my_epptr);
	     pbump((int) newpos);
	     return (int32_t) newpos;
	  }
    };

    // return the length of the current data object.
    int32_t length(void) {
        std::streampos cpos = currPos();
	return (cpos == std::streampos(EOF)) ? (int32_t) std::streampos(blen()) : (int32_t) cpos;
    };

    // move a portion of the buffer from one spot 
    // to another.
    void move(std::streampos dest, std::streampos src, size_t length) {
        memmove(pbase()+dest, pbase()+src, length);
    };

    size_t size(void) { return length(); };
    unsigned char* data(void) { return (unsigned char*) pbase(); };
    
    // function for writing a buffer to the stream
    asn1ostream& write(unsigned char *data, int32_t len) {
        while(len-- > 0 && this->good()) {
            this->put(*data++);
        }
	return (*this);
    };

    // length of the whole message 
    size_t msgLength(void) {
        return (epptr() - pbase()); 
    }

    // Whether the stream needs encryption or not
    bool encrypt_:1;

    char *m_stringBuff;

    // low-level output functions
    asn1ostream& operator<<(uint32_t);    
    asn1ostream& operator<<(int32_t);
    asn1ostream& operator<<(uint64_t);
    asn1ostream& operator<<(char*);    // display string (only).
    asn1ostream& operator<<(unsigned char c) {
        this->put(c); return (*this);
    };

    // stream reset read and write operators.
    void reset(void);
};

//
// asn1 (input) stream 
//
class asn1istream : public std::streambuf,
		    public virtual std::ios
{
    // a position pointer remembers a point in
    // the stream buffer. It is used in the asn1ObjContext
    // stack to record the starting points of ASN1
    // context and sequences.
    std::streampos asn1ObjContext[10];
    int32_t contextIndex;

    void pushASNContext(std::streampos p) {
        if (contextIndex < 10) {
	    asn1ObjContext[contextIndex] = p;
	    contextIndex++;
	} else {
	    std::ios::setstate(std::ios::badbit|std::ios::failbit);
	}
    }

    std::streampos popASNContext(void) {
      if (contextIndex > 0) {
	  contextIndex--;
	  return asn1ObjContext[contextIndex];
      } else {
	  std::ios::setstate(std::ios::badbit|std::ios::failbit);
	  return 0;
      }
    }

    std::streampos peekASNContext(void) {
       return asn1ObjContext[contextIndex-1];
    }

    // a stack containing asn object lengths
    // that correspond with the ASN context stack.
    size_t currRecLength[10];
    int32_t   recIndex;

    void pushASNSequenceLen(size_t len) {
        if (recIndex < 10) {
	   currRecLength[recIndex] = len;
	   recIndex++;
	} else {
	  std::ios::setstate(std::ios::badbit|std::ios::failbit);
	}
    }

    size_t popASNSequenceLen(void) {
        if (recIndex > 0) {
	    recIndex--;
	    return currRecLength[recIndex];
	} else {
	   std::ios::setstate(std::ios::badbit|std::ios::failbit);
	   return 0; 
	}
    }

    size_t peekASNSequenceLen(void) {
        return currRecLength[recIndex-1];
    }
    
    int blen() const { return (int) (egptr() - eback()); }

public:
    //
    // currPos - return the current position in the read buffer
    //           as a integer stream position.
    //           (current read pointer - buffer base)
    //
    std::streampos currPos(void) {
        std::streampos offset = gptr() - eback();
	return offset;
    };

    friend asn1istream& _endObj(asn1istream&);
    friend asn1istream& readSequence(asn1istream&);
    friend asn1istream& readSecurityParams(asn1istream&);
    friend int32_t readContext(asn1istream&);
    friend bool isSequenceEnd(asn1istream&);

    int32_t ReadOk(void) const { return ! bad(); };

    // passed buffer
    asn1istream(unsigned char *b, int32_t len);

    // Destructor
    ~asn1istream(void) {};

    // get a character.
    int32_t get(void) {
        int32_t c=0;
	if ((c = this->std::streambuf::sbumpc()) == EOF) {
	    this->std::ios::setstate(std::ios::eofbit|std::ios::failbit);
	}
	return c;
    };

    // where would we be with out an unget?
    void unget(char c) {
        /* don't unget on corrupt streams. */
        if (!bad()) {
	   if (eof()) {
	       clear();
	   }
	   std::streambuf::sputbackc(c);
	}
    };


    // move the read pointer to "newpos" in the
    // input buffer.
    int32_t seek(std::streampos newpos) {
        std::streampos pos = seekpos(newpos, std::ios::in);
	if (pos == std::streampos(EOF)) {
	    std::ios::setstate(std::ios::badbit|std::ios::failbit);
	}
	return (int32_t) pos;
    };


    // for reading a buffer from the stream
    int32_t read(unsigned char *, size_t, uint32_t);

    // low-level input functions
    asn1istream& operator>>(uint32_t&);    
    asn1istream& operator>>(int32_t&);
    asn1istream& operator>>(uint64_t&);
    asn1istream& operator>>(char*);  
    asn1istream& operator>>(unsigned char& c) {
        unsigned char c1 = this->get();
	if (! eof()) { c = c1; }
	return (*this);
    };

    // stream reset read and write operators.
    void reset(void);
    
    // Function for rewriting a buffer into the asn1istream
    asn1istream& rewrite(const char *data, int32_t len);

    size_t size(void) {
	std::streampos cpos = currPos();
	return (cpos == std::streampos(EOF)) ? std::streampos(blen()) : cpos;
    }	    
    unsigned char* data(void) { return (unsigned char*) eback(); };

    // length of the whole message as received on the wire
    size_t msgLength(void) {
	return blen();
    }
};

//
//  The number of bytes that are reserved for 
//  encoding asn.1 object lengths when writing asn.1
//  sequences. The length is made up of one tag/length
//  byte followed by 2 bytes that contain the following
//  object's length. 2 bytes enables the encoding of
//  objects up to 64 kb in length.
//
//  IPv4 packets have a max size of 64 kb.
//  
#define ASN_LONGLENGTH_SIZE 3

// output manipulators
typedef asn1ostream& (*Aomanip)(asn1ostream&);

inline asn1ostream& operator<<(asn1ostream& s, Aomanip f)
{
    return f(s);
}

// input manipulators
typedef asn1istream& (*Aimanip)(asn1istream&);

inline asn1istream& operator>>(asn1istream& s, Aimanip f)
{
    return f(s);
}

// sequence operators
asn1ostream& beginSequence(asn1ostream&);
asn1ostream& beginSecurityParams(asn1ostream&);
asn1ostream& endSequence(asn1ostream&);
asn1ostream& endSecurityParams(asn1ostream&);
asn1ostream& cancelSequence(asn1ostream&);

asn1istream& readSequence(asn1istream&);
asn1istream& readSecurityParams(asn1istream&);
asn1istream& endSequence(asn1istream&);
asn1istream& endContext(asn1istream&);
bool isSequenceEnd(asn1istream&);
			   
// manipulators with arguments
template<class T> class asn1omanip {
    asn1ostream& (*_f)(asn1ostream&, T);
    T _a;
public:
    asn1omanip(asn1ostream& (*f)(asn1ostream&, T), T a) : _f(f), _a(a) {};
    ~asn1omanip() {};
    friend asn1ostream& operator<<(asn1ostream& s, const asn1omanip& m)
	{ return m._f(s, m._a); }
};

extern asn1ostream& _null(asn1ostream& s, unsigned char type);
inline asn1omanip<unsigned char> asnNull(unsigned char type)
                { return asn1omanip<unsigned char>(_null, type);}

// 
//
extern asn1ostream& _beginPrimitive(asn1ostream& s, unsigned char type);
inline asn1omanip<unsigned char> beginPrimitive(unsigned char type)
                { return asn1omanip<unsigned char>(_beginPrimitive, type);}


// read the start of an ASN1 context.
//
int32_t readContext(asn1istream& s);
//
// write the start of the ASN1 context.
//
extern asn1ostream& _beginContext(asn1ostream& s, unsigned char type);
inline asn1omanip<unsigned char> beginContext(unsigned char type)
                { return asn1omanip<unsigned char>(_beginContext, type);}

// adding an endContext manipulator to the asn1 output
// stream closes the currently open context.
asn1ostream& endContext(asn1ostream&);

//
// Converts and integer into an ASN.1 length
extern asn1ostream& _Len(asn1ostream& s, int32_t len);
inline asn1omanip<int32_t> Len(int len) {return asn1omanip<int32_t>(_Len, len);}

// returns the length of the following object in bytes.
//
size_t getLength(asn1istream& s);
#endif // _asn1stream_h_
